import sys
from pathlib import Path

from qtpy.QtGui import QIcon, QKeySequence
from qtpy.QtWidgets import QAction, QStyle, QMessageBox, QFileDialog, QVBoxLayout
from pyqode.python.folding import PythonFoldDetector
from qtpy.QtWidgets import QWidget, QApplication
from qtpy.QtCore import Qt
from pyqode.core import panels, api, modes
from pyqode.python import widgets, panels as pypanels, modes as pymodes
from pyqode.python.backend import server

from tsuchinoko import assets
from tsuchinoko.examples import server_demo
from pyqode.python.backend.workers import defined_names
from pyqode.python import managers as pymanagers

import inspect

from tsuchinoko.widgets.debugmenubar import DebuggableMenuBar


class PythonEditor(widgets.PyCodeEditBase):
    def __init__(self):
        super(PythonEditor, self).__init__()

        # starts the default pyqode.python server (which enable the jedi code
        # completion worker).
        suffix = Path(sys.executable).suffix
        bootstrap_exe = (Path(sys.executable).parent / 'tsuchinoko_bootstrap').with_suffix(
            suffix if suffix == '.exe' else '')
        self.backend.start(server.__file__, interpreter=str(bootstrap_exe))

        # some other modes/panels require the analyser mode, the best is to
        # install it first
        self.modes.append(modes.OutlineMode(defined_names))

        # --- core panels
        self.panels.append(panels.FoldingPanel())
        self.panels.append(panels.LineNumberPanel())
        self.panels.append(panels.CheckerPanel())
        self.panels.append(panels.SearchAndReplacePanel(),
                           panels.SearchAndReplacePanel.Position.BOTTOM)
        self.panels.append(panels.EncodingPanel(), api.Panel.Position.TOP)
        # add a context menu separator between editor's
        # builtin action and the python specific actions
        self.add_separator()

        # --- python specific panels
        self.panels.append(pypanels.QuickDocPanel(), api.Panel.Position.BOTTOM)

        # --- core modes
        self.modes.append(modes.CaretLineHighlighterMode())
        self.modes.append(modes.CodeCompletionMode())
        self.modes.append(modes.ExtendedSelectionMode())
        self.modes.append(modes.FileWatcherMode())
        self.modes.append(modes.OccurrencesHighlighterMode())
        self.modes.append(modes.RightMarginMode())
        self.modes.append(modes.SmartBackSpaceMode())
        self.modes.append(modes.SymbolMatcherMode())
        self.modes.append(modes.ZoomMode())
        # self.modes.append(modes.PygmentsSyntaxHighlighter(self.document()))

        # ---  python specific modes
        self.modes.append(pymodes.CommentsMode())
        self.modes.append(pymodes.CalltipsMode())
        self.modes.append(pymodes.PyFlakesChecker())
        self.modes.append(pymodes.PEP8CheckerMode())
        self.modes.append(pymodes.PyAutoCompleteMode())
        self.modes.append(pymodes.PyAutoIndentMode())
        self.modes.append(pymodes.PyIndenterMode())
        self.modes.append(pymodes.PythonSH(self.document()))
        self.syntax_highlighter.fold_detector = PythonFoldDetector()

        self.syntax_highlighter.color_scheme = api.ColorScheme('darcula')
        self.save_on_focus_out = True


class ServerEditor(QWidget):
    def __init__(self, main_window):
        super().__init__()

        self.main_window = main_window
        self.editor = PythonEditor()

        menubar = DebuggableMenuBar()
        file_menu = menubar.addMenu("&File")
        file_menu.addAction('&New', self.new)
        open_action = QAction(self.style().standardIcon(QStyle.SP_DirOpenIcon), 'Open...', parent=file_menu)
        save_action = QAction(self.style().standardIcon(QStyle.SP_DialogSaveButton), 'Save',
                                   parent=file_menu)
        save_action.setShortcut(QKeySequence(Qt.CTRL | Qt.Key_S))
        save_as_action = QAction(self.style().standardIcon(QStyle.SP_DialogSaveButton), 'Save As...',
                                   parent=file_menu)
        server_menu = menubar.addMenu("&Server")
        start_server = QAction(self.style().standardIcon(QStyle.SP_MediaPlay), 'Start Server', parent=server_menu)
        stop_server = QAction(self.style().standardIcon(QStyle.SP_MediaStop), 'Stop Server', parent=server_menu)

        file_menu.addAction(open_action)
        file_menu.addAction(save_action)
        file_menu.addAction(save_as_action)
        file_menu.addAction('E&xit', self.exit)
        server_menu.addAction(start_server)
        server_menu.addAction(stop_server)

        open_action.triggered.connect(self.open)
        save_action.triggered.connect(self.save)
        save_as_action.triggered.connect(self.save_as)
        start_server.triggered.connect(self.start_server)
        stop_server.triggered.connect(self.stop_server)

        self.setWindowTitle('Tsuchinoko Server Editor')
        self.setWindowIcon(QIcon(assets.path('tsuchinoko.png')))
        self.resize(1700, 1000)

        self.has_unsaved_changes = False
        self.editor.textChanged.connect(self._set_has_changes)

        QApplication.instance().aboutToQuit.connect(self.close_by_quit)  # TODO: use this approach in Xi-cam

        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        layout.addWidget(menubar)
        layout.addWidget(self.editor)
        self.setLayout(layout)

        self.new()
        self.editor.setPlainText(inspect.getsource(server_demo))

    def _set_has_changes(self):
        self.has_unsaved_changes = True

    def _prompt_unsaved_changes(self, buttons=QMessageBox.Yes | QMessageBox.No | QMessageBox.Cancel):
        result = None
        if self.has_unsaved_changes:
            result = QMessageBox.question(self,
                                          'Unsaved changes',
                                          f"Do you want to save changes to {Path(self.editor.file.path).name or 'Untitled'}?",
                                          buttons=QMessageBox.StandardButtons(buttons),
                                          defaultButton=QMessageBox.Yes)

            if not result:
                raise ValueError("WHAT?!")

        return result

    def _prompt_save_path(self):
        name, filter = QFileDialog.getSaveFileName(filter=("Python (*.py)"))
        if not name:
            return False
        return name

    def new(self):
        if self._prompt_unsaved_changes() == QMessageBox.Cancel:
            return False
        self.editor.file = pymanagers.PyFileManager(self.editor)
        self.editor.setPlainText('')
        self.has_unsaved_changes = False
        return True

    def open(self):
        if self._prompt_unsaved_changes() == QMessageBox.Cancel:
            return False
        name, filter = QFileDialog.getOpenFileName(filter=("Python (*.py)"))
        if not name:
            return False
        self.editor.file = pymanagers.PyFileManager(self.editor)
        self.editor.file.open(name)
        self.has_unsaved_changes = False
        return True
        # with open(name, 'r') as f:
        #     state = f.read()
        # self.editor.setPlainText(state)

    def save(self):
        path = self.editor.file.path
        return self.save_as(path)

    def save_as(self, path=None):
        if not path:
            path = self._prompt_save_path()
            if not path: return False

        # state = self.editor.toPlainText()
        # with open(path, 'w') as f:
        #     f.write(state)

        self.editor.file.save(path)
        self.has_unsaved_changes = False
        # self.path = path
        return True

    def exit(self, buttons=QMessageBox.Yes | QMessageBox.No | QMessageBox.Cancel):
        if self._prompt_unsaved_changes(buttons) == QMessageBox.Cancel:
            return False

        self.editor.file.close()
        self.editor.backend.stop()
        self.has_unsaved_changes = False
        self.close()

        return True

    def closeEvent(self, event):
        if self.exit():
            event.accept()  # let the window close
        else:
            event.ignore()

    def close_by_quit(self):
        self.exit(buttons=QMessageBox.Yes | QMessageBox.No)

    def start_server(self):
        if self._prompt_unsaved_changes(buttons=QMessageBox.Yes | QMessageBox.Cancel) == QMessageBox.Cancel:
            return False
        self.save()
        self.main_window.start_server(path=self.editor.file.path)

    def stop_server(self):
        self.main_window.stop_server(confirm=True)



"""
Pulls data from 3D hyperspectral h5 data block; performs analysis using reference spectra
"""

import os
import time

import h5py
import numpy as np

from tsuchinoko.adaptive.gpCAM_in_process import GPCAMInProcessEngine
from tsuchinoko.core import ZMQCore
from tsuchinoko.execution.threaded_in_process import ThreadedInProcessEngine


def val2ind(val, an_array):
    return np.argmin(abs(an_array - val), axis=0)


class ir_getdata:

    def __init__(self, fileName, ref_path, print_dir=False, dataName=r'./gpdata/Test'):
        h5_groups = []
        epsilon, b = 0.4, 0.29
        with h5py.File(fileName, 'r') as f:
            root_name = list(f.keys())[0]
            print(root_name)
            h = f[root_name]  # folder name
            self.xy = f[root_name + '/data/xy'][:, :]
            self.wav = f[root_name + '/data/wavenumbers'][:]
            self.N_w = len(self.wav)
            self.data_list = f[root_name + '/data/spectra'][:, :]
            self.data_cube = f[root_name + '/data/image/image_cube'][:, :, :]
            self.data_list = -np.log10(self.data_list / 100 + epsilon) + b
            self.data_cube = -np.log10(self.data_cube / 100 + epsilon) + b
            h.visit(h5_groups.append)
            if print_dir:
                print(*h5_groups, sep='\n')
        self.n_spec = 0
        self.cum_spectra = np.zeros((0, self.N_w))
        path = os.path.dirname(dataName)
        self.cum_specName = os.path.join(path, 'all_specs.npy')
        self.ref_spec = np.loadtxt(ref_path, delimiter=',')
        print(f'ref_spec: {os.path.basename(ref_path)}')

        # get X, Y coordinates
        self.X = np.unique(self.xy[:, 0])
        self.Y = np.unique(self.xy[:, 1])

    def get_spec(self, pos):
        row, col = [], []
        pos = pos.reshape(-1, 2)
        for i in range(len(pos)):
            row.append(val2ind(pos[i, 1], self.Y))  # align y coordinate to get row
            col.append(val2ind(pos[i, 0], self.X))  # align x coordinate to get col
        spec = self.data_cube[row, col, :]
        self.cum_spectra = np.append(self.cum_spectra, spec.reshape(1, -1), axis=0)
        np.save(self.cum_specName, self.cum_spectra)
        self.n_spec += 1

    def alignWithRef(self, spec, ref):
        if ref.ndim == 1:
            return spec, ref
        elif ref.ndim == 2:
            pos2, pos1 = val2ind(self.wav, ref[-1, 0]) + 1, val2ind(self.wav, ref[0, 0])
            if spec.ndim == 1: spec = spec.reshape(1, -1)
            if len(self.wav[pos1: pos2]) == ref.shape[0]:
                return spec[:, pos1: pos2], ref[:, 1]
            elif len(self.wav[pos1: pos2]) == ref[::2, :].shape[0]:
                return spec[:, pos1: pos2], ref[::2, 1]
            elif len(self.wav[pos1: pos2][::4]) == ref.shape[0]:
                return spec[:, pos1: pos2: 4], ref[:, 1]
            elif len(self.wav[pos1: pos2][::4]) == ref[::2, :].shape[0]:
                return spec[:, pos1: pos2: 4], ref[::2, 1]
            else:
                print("Length doesn't match")
                return None

    def gp_getCorrData(self, idx):
        spec = self.cum_spectra[idx, :]
        spec, ref = self.alignWithRef(spec, self.ref_spec)
        return np.corrcoef(spec, ref[None, :])[0, 1]


def instrument_synthetic(data):
    # load last saved data
    variance = 1e-3

    if (ir_data.n_spec == 0) and os.path.exists(ir_data.cum_specName):
        ir_data.cum_spectra = np.load(ir_data.cum_specName)
        ir_data.n_spec = ir_data.cum_spectra.shape[0]

    for idx_data in range(len(data)):
        if data[idx_data]["measured"] == True: continue
        x1 = data[idx_data]["position"][0]
        x2 = data[idx_data]["position"][1]
        pos = np.array([x1, x2])
        ir_data.get_spec(pos)

    for idx_data in range(len(data)):
        if data[idx_data]["measured"] == True: continue
        data[idx_data]["value"] = ir_data.gp_getCorrData(-1)  # Modified for single-point operation idx_data)
        data[idx_data]["variance"] = variance
        data[idx_data]["measured"] = True
        data[idx_data]["time stamp"] = time.time()
    return data


fileName = 'E:\\data\\IR-gpcam\\20180216r_S1_tr_area1.h5'
# ref_path = 'E:\\data\\IR-gpcam\\Lipopolysaccharides.CSV'
ref_path = 'E:\\data\\IR-gpcam\\black_carbons_on_minerals.CSV'
dataName = 'E:\\data\\IR-gpcam\\'
ir_data = ir_getdata(fileName, ref_path, dataName=dataName)

if __name__ == '__main__':

    # Define a function to measure target positions
    def measure_target(target):
        data = instrument_synthetic([{'position': target, 'measured': False}])
        return data[0]['value'], data[0]['variance']

    # Define a gpCAM adaptive engine with initial parameters
    adaptive = GPCAMInProcessEngine(dimensionality=2,
                                    parameter_bounds=[[-20590.4, -20010.4], [3394.0, 3894.0]],
                                    hyperparameters=[1, 1, 1],
                                    hyperparameter_bounds=[[0.0001, 10], [0.001, 1e4], [0.001, 1e4]])

    # Define an execution engine with the measurement function
    execution = ThreadedInProcessEngine(measure_target)

    # Construct and start a core server
    core = ZMQCore()
    core.set_adaptive_engine(adaptive)
    core.set_execution_engine(execution)
    core.main()



# -*- coding: utf-8 -*-
""" rwlock.py
    A class to implement read-write locks on top of the standard threading
    library.
    This is implemented with two mutexes (threading.Lock instances) as per this
    wikipedia pseudocode:
    https://en.wikipedia.org/wiki/Readers%E2%80%93writer_lock#Using_two_mutexes
    Code written by Tyler Neylon at Unbox Research.
    This file is public domain.
"""


# _______________________________________________________________________
# Imports

from contextlib import contextmanager
from threading  import Lock


# _______________________________________________________________________
# Class

class RWLock(object):
    """ RWLock class; this is meant to allow an object to be read from by
        multiple threads, but only written to by a single thread at a time. See:
        https://en.wikipedia.org/wiki/Readers%E2%80%93writer_lock
        Usage:
            from rwlock import RWLock
            my_obj_rwlock = RWLock()
            # When reading from my_obj:
            with my_obj_rwlock.r_locked():
                do_read_only_things_with(my_obj)
            # When writing to my_obj:
            with my_obj_rwlock.w_locked():
                mutate(my_obj)
    """

    def __init__(self):

        self.w_lock = Lock()
        self.num_r_lock = Lock()
        self.num_r = 0

    # ___________________________________________________________________
    # Reading methods.

    def r_acquire(self):
        self.num_r_lock.acquire()
        self.num_r += 1
        if self.num_r == 1:
            self.w_lock.acquire()
        self.num_r_lock.release()

    def r_release(self):
        assert self.num_r > 0
        self.num_r_lock.acquire()
        self.num_r -= 1
        if self.num_r == 0:
            self.w_lock.release()
        self.num_r_lock.release()

    @contextmanager
    def r_locked(self):
        """ This method is designed to be used via the `with` statement. """
        try:
            self.r_acquire()
            yield
        finally:
            self.r_release()

    # ___________________________________________________________________
    # Writing methods.

    def w_acquire(self):
        self.w_lock.acquire()

    def w_release(self):
        self.w_lock.release()

    @contextmanager
    def w_locked(self):
        """ This method is designed to be used via the `with` statement. """
        try:
            self.w_acquire()
            yield
        finally:
            self.w_release()


import numbers
import pickle
import time
from typing import Tuple, List

import numpy as np
import zmq
from loguru import logger

from . import Engine
from ..adaptive.gpCAM_in_process import GPCAMInProcessEngine

SLEEP_FOR_AGENT_TIME = .1
SLEEP_FOR_TSUCHINOKO_TIME = .1
FORCE_KICKSTART_TIME = 5


class BlueskyAdaptiveEngine(Engine):
    """
    A `tsuchinoko.execution.Engine` that sends targets to Blueskly-Adaptive and receives back measured data.
    """

    suggest_blacklist = ["x_data",
                         "y_data",
                         "noise_variances",
                         "init_hyperparameters"]  # keys with ragged state

    def __init__(self, adaptive_engine:GPCAMInProcessEngine, host: str = '127.0.0.1', port: int = 5557):
        """

        Parameters
        ----------
        host
            A host address target for the zmq socket.
        port
            The port used for the zmq socket.
        """
        super(BlueskyAdaptiveEngine, self).__init__()

        self.adaptive_engine = adaptive_engine
        self.position = None
        self.context = None
        self.socket = None
        self.host = host
        self.port = port
        self.setup_socket()
        self._last_targets_sent = None
        # Lock sending new points until at least one from the previous list is measured
        self.has_fresh_points_on_server = False

    def setup_socket(self):
        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.PAIR)

        # Attempt to bind, retry every second if fails
        while True:
            try:
                self.socket.bind(f"tcp://{self.host}:{self.port}")
            except zmq.ZMQError as ex:
                logger.info(f'Unable to bind to tcp://{self.host}:{self.port}. Retrying in 1 second...')
                logger.exception(ex)
                time.sleep(1)
            else:
                logger.info(f'Bound to tcp://{self.host}:{self.port}.')
                break

    def update_targets(self, targets: List[Tuple]):
        if self.has_fresh_points_on_server:
            time.sleep(SLEEP_FOR_AGENT_TIME)  # chill if the Agent hasn't measured any points from the previous list
        else:
            # checkpoint optimizer state
            gpcam_state = self.adaptive_engine.optimizer.__getstate__()

            # sanitize state
            for key in self.suggest_blacklist:
                if gpcam_state[key] is None:
                    gpcam_state[key] = np.array([])
            sanitized_gpcam_state = dict(
                **{key if key not in self.suggest_blacklist else f"STATEDICT-{key}":
                       np.asarray(val) if isinstance(val, (list, tuple)) else val
                   for key, val in gpcam_state.items()
                   if val is not None}  # event-model doesn't like Nones
                )

            # send targets to TsuchinokoAgent
            self.has_fresh_points_on_server = self.send_payload({'candidate': targets,
                                                                 'optimizer': sanitized_gpcam_state})
            self._last_targets_sent = targets

    def get_measurements(self) -> List[Tuple]:
        new_measurements = []
        # get newly completed measurements from bluesky-adaptive; repeat until buffered payloads are exhausted
        while True:
            try:
                payload = self.recv_payload(flags=zmq.NOBLOCK)
            except zmq.ZMQError:
                break
            else:
                assert 'target_measured' in payload
                x, (y, v) = payload['target_measured']
                new_measurements.append((x, y, v, {}))
                # stash the last position measured as the 'current' position of the instrument
                self.position = x
        if new_measurements:
            self.has_fresh_points_on_server = False
        return new_measurements

    def get_position(self) -> Tuple:
        # return last measurement position received from bluesky-adaptive
        return self.position

    def send_payload(self, payload: dict):
        logger.info(f'message: {payload}')
        try:
            self.socket.send(pickle.dumps(payload), flags=zmq.NOBLOCK)
        except zmq.error.Again:
            return False
        return True

    def recv_payload(self, flags=0) -> dict:
        payload_response = pickle.loads(self.socket.recv(flags=flags))
        logger.info(f'response: {payload_response}')
        # if the returned message is the kickstart message, resend the last targets sent and check for more payloads
        if payload_response == {'send_targets': True}:
            self.has_fresh_points_on_server = False
            self.update_targets(self._last_targets_sent)
            payload_response = self.recv_payload(flags)
        return payload_response



from abc import ABC, abstractmethod
from typing import Tuple, List


class Engine(ABC):
    """
    The Execution Engine base class. This component is generally to be responsible for measuring targets.
    """

    @abstractmethod
    def update_targets(self,  targets: List[Tuple]):
        """
        Updates the list of measurement targets. Generally new targets will take precedence, and will be stashed for
        measurement in the background.

        Parameters
        ----------
        targets: list
            A list of new measure target positions.
        """
        ...

    @abstractmethod
    def get_position(self) -> Tuple:
        """
        Returns the current position in the target domain.

        Returns
        -------
        current_position: tuple
            The current position in the target domain

        """
        ...

    @abstractmethod
    def get_measurements(self) -> List[Tuple]:
        """
        Returns new measurements made since its last call. Generally measured values are accumulated by a background thread,
        then returned and cleared here.

        Returns
        -------
        position: tuple
            The measured target position.
        value: float
            The value at that position.
        variance: float
            The variance associated with the measurement of that value.
        metrics: dict
            Any non-standard metric values to be used for visualization at the client.
        """
        ...



# This module is separated to limit imports done in child processes
from multiprocessing import Queue
import time


targets = Queue()
measurements = Queue()


def measure(position):
    targets.put(position)
    return measurements.get()


def set_global(*args):
    global targets, measurements
    targets, measurements = args
    # print('ID:', id(targets))
    # print(targets, measurements)



import time
from collections import defaultdict
from functools import partial
from pickle import UnpicklingError
from queue import Queue, Empty
import subprocess
from signal import SIGINT
from typing import Any, Type, Union
import sys
from pathlib import Path

from tsuchinoko.utils.dependencies import check_dependencies
from tsuchinoko.widgets.debugmenubar import DebuggableMenuBar
from tsuchinoko.widgets.server_editor import ServerEditor

try:
    from yaml import CLoader as Loader, CDumper as Dumper, dump, load
except ImportError:
    from yaml import Loader, Dumper

import zmq
from zmq.error import ZMQError, Again
import numpy as np
from qtpy.QtGui import QIcon
from loguru import logger
from pyqtgraph import mkBrush, mkPen, HistogramLUTWidget, PlotItem
from pyqtgraph.dockarea import DockArea
from qtmodern.styles import dark
from qtpy.QtWidgets import QMainWindow, QApplication, QHBoxLayout, QWidget, QMenuBar, QAction, QStyle, QFileDialog, QDialog, QMessageBox

from tsuchinoko.assets import path
from tsuchinoko.adaptive import Data
from tsuchinoko.core import CoreState
from tsuchinoko.core.messages import PauseRequest, StartRequest, GetParametersRequest, SetParameterRequest, \
    PartialDataRequest, FullDataRequest, StopRequest, Message, StateRequest, StateResponse, GetParametersResponse, \
    FullDataResponse, PartialDataResponse, MeasureRequest, \
    ConnectRequest, ConnectResponse, PushDataRequest, ExceptionResponse, PullGraphsRequest, GraphsResponse, \
    ReplayRequest, ExitRequest, PushGraphsRequest, SetComputeMetricsRequest
from tsuchinoko.graphics_items.clouditem import CloudItem
from tsuchinoko.graphics_items.indicatoritem import BetterCurveArrow
from tsuchinoko.graphics_items.mixins import ClickRequester, request_relay, ClickRequesterPlot
from tsuchinoko.utils.threads import QThreadFutureIterator, invoke_as_event
from tsuchinoko.widgets.displays import Log, Configuration, GraphManager, StateManager


class ImageViewBlend(ClickRequester):
    pass


class MainWindow(QMainWindow):
    def __init__(self, core_address='localhost'):
        super(MainWindow, self).__init__()

        menubar = DebuggableMenuBar()
        file_menu = menubar.addMenu("&File")
        file_menu.addAction('&New...', self.new_data)
        open_data_action = QAction(self.style().standardIcon(QStyle.SP_DirOpenIcon), 'Open data...', parent=file_menu)
        open_parameters_action = QAction(self.style().standardIcon(QStyle.SP_DirOpenIcon), 'Open parameters...', parent=file_menu)
        save_data_action = QAction(self.style().standardIcon(QStyle.SP_DialogSaveButton), 'Save data as...', parent=file_menu)
        save_parameters_action = QAction(self.style().standardIcon(QStyle.SP_DialogSaveButton), 'Save parameters as...', parent=file_menu)
        open_server_editor = QAction(self.style().standardIcon(QStyle.SP_DirOpenIcon), 'Open Server Editor', parent=file_menu)
        file_menu.addAction(open_data_action)
        file_menu.addAction(open_parameters_action)
        file_menu.addAction(save_data_action)
        file_menu.addAction(save_parameters_action)
        file_menu.addAction(open_server_editor)
        file_menu.addAction('E&xit', self.close)
        self.setMenuBar(menubar)

        demo_menu = menubar.addMenu("&Demo")
        demo_menu.addAction('Start &Simple Demo', partial(self.start_demo, 'server_demo'))
        demo_menu.addAction('Start &Adaptive Demo', partial(self.start_demo, 'adaptive_demo')).setEnabled(check_dependencies(['adaptive']))
        demo_menu.addAction('Start &Grid Scan Demo', partial(self.start_demo, 'grid_demo'))
        demo_menu.addAction('Start &High Dimensionality Demo',  partial(self.start_demo, 'high_dimensionality_server_demo')).setEnabled(check_dependencies(['perlin-noise']))
        demo_menu.addAction('Start &Multi Task Demo',  partial(self.start_demo, 'multi_task_server_demo'))
        demo_menu.addAction('Start &Quad Tree Demo',  partial(self.start_demo, 'quadtree_demo'))
        demo_menu.addAction('Start &Bluesky Demo',  partial(self.start_demo, 'server_demo_bluesky'))

        save_data_action.triggered.connect(self.save_data)
        open_data_action.triggered.connect(self.open_data)
        save_parameters_action.triggered.connect(self.save_parameters)
        open_parameters_action.triggered.connect(self.open_parameters)
        open_server_editor.triggered.connect(self.open_server_editor)

        self.setWindowTitle('Tsuchinoko')
        self.setWindowIcon(QIcon(path('tsuchinoko.png')))
        self.resize(1700, 1000)

        self.log_widget = Log()
        self.configuration_widget = Configuration()
        self.state_manager_widget = StateManager()
        self.graph_manager_widget = GraphManager()

        self.dock_area = DockArea()
        self.setCentralWidget(self.dock_area)

        for position, w, *relaltive_to in [('bottom', self.graph_manager_widget),
                                           ('bottom', self.log_widget, self.graph_manager_widget),
                                           ('right', self.configuration_widget, self.graph_manager_widget),
                                           ('top', self.state_manager_widget, self.configuration_widget),
                                           ]:
            self.dock_area.addDock(w, position, *relaltive_to)

        dark(QApplication.instance())

        self.context = zmq.Context()
        self.socket = None
        self.core_address = core_address
        self.init_socket()

        self.state_manager_widget.sigPause.connect(self.pause)
        self.state_manager_widget.sigStart.connect(self.start)
        self.state_manager_widget.sigStop.connect(self.stop)
        self.state_manager_widget.sigReplay.connect(self.replay)
        self.state_manager_widget.sigSetComputeMetrics.connect(self.set_compute_metrics)
        self.configuration_widget.sigPushParameter.connect(self.set_parameter)
        self.configuration_widget.sigRequestParameters.connect(self.request_parameters)
        self.graph_manager_widget.sigPush.connect(self.push_graph)
        request_relay.sigRequestMeasure.connect(self.request_measure)

        self.update_thread = QThreadFutureIterator(self.update, finished_slot=self.close_zmq, name='tsuchinoko-update')
        self.update_thread.start()

        self.data: Data = Data()
        self.last_data_size = None
        self.callbacks = defaultdict(list)

        self.subscribe(self.state_manager_widget.update_state, StateResponse)
        self.subscribe(self.state_manager_widget.update_state, ConnectResponse)
        self.subscribe(self.configuration_widget.update_parameters, GetParametersResponse, invoke_as_event=True)
        self.subscribe(partial(self._data_callback, response_type='full'), FullDataResponse)
        self.subscribe(partial(self._data_callback, response_type='partial'), PartialDataResponse)
        self.subscribe(self.refresh_state, ConnectResponse)
        self.subscribe(self.log_widget.log_exception, ExceptionResponse)
        self.subscribe(self.set_graphs, GraphsResponse, invoke_as_event=True)

        self._server = None

    def init_socket(self):
        if self.socket:
            logger.debug("Closing socket")
            self.socket.close()

        #  Socket to talk to server
        logger.info("Connecting to core server…")
        self.socket = self.context.socket(zmq.REQ)
        self.socket.setsockopt(zmq.LINGER, 5)
        self.socket.connect(f"tcp://{self.core_address}:5555")
        self.socket.RCVTIMEO = 5000
        self.message_queue = Queue()

    def try_connect(self):
        self.message_queue.put(ConnectRequest())

    def get_state(self):
        self.message_queue.put(StateRequest())

    def pause(self):
        self.message_queue.put(PauseRequest())

    def start(self):
        self.message_queue.put(StartRequest())

    def stop(self):
        self.message_queue.put(StopRequest())

    def replay(self):
        message = ReplayRequest(self.data.positions, self.data.measurements)
        self.message_queue.put(StopRequest())
        self.message_queue.put(message)
        self.message_queue.put(StartRequest())

    def set_compute_metrics(self, value):
        self.message_queue.put(SetComputeMetricsRequest(value))

    def request_measure(self, pos):
        self.message_queue.put(MeasureRequest(pos))

    def request_parameters(self):
        self.message_queue.put(GetParametersRequest())

    def push_graph(self, graph):
        self.message_queue.put(PushGraphsRequest([graph]))

    def set_parameter(self, child_path: str, value: Any):
        if child_path:
            self.message_queue.put(SetParameterRequest(child_path, value))

    def update(self):
        self.data = Data()
        self.last_data_size = 0

        while True:
            yield
            request = None

            if self.state_manager_widget.state == CoreState.Stopping:
                self.last_data_size = 0
                self.data = Data()

            if self.state_manager_widget.state == CoreState.Connecting:
                self.try_connect()

            if self.state_manager_widget.state in [CoreState.Pausing, CoreState.Starting, CoreState.Resuming, CoreState.Resuming, CoreState.Stopping]:
                self.get_state()

            try:
                request = self.message_queue.get(timeout=.2)
            except Empty:
                if self.state_manager_widget.state == CoreState.Running:
                    if self.data:
                        request = PartialDataRequest(len(self.data))
                    else:
                        request = FullDataRequest()

            if not request:
                self.get_state()
                continue
            else:
                logger.info(f'request: {request}')

            try:
                self.socket.send_pyobj(request)
                response = self.socket.recv_pyobj()
            except (ZMQError, Again) as ex:
                logger.warning(f'Unable to connect to core server at {self.core_address}...')
                time.sleep(1)
                # logger.exception(ex)
                if self.context:
                    self.init_socket()
                self.data = Data()  # wipeout data and get a full update next time
                self.last_data_size = 0
                self.state_manager_widget.update_state(CoreState.Connecting, True)
            except UnpicklingError as ex:
                logger.exception(ex)
                logger.critical('The above error prevented unpacking data from the server.')
            else:
                logger.info(f'response: {response}')
                if not response:
                    self.get_state()
                else:
                    if self.state_manager_widget.state == CoreState.Connecting:
                        logger.critical(f'Successfully connected to server at {self.core_address}.')
                    for callback, as_event in self.callbacks[type(response)]:
                        if as_event:
                            invoke_as_event(callback, *response.payload)
                        else:
                            callback(*response.payload)

    def _data_callback(self, data_payload, last_data_size=None, response_type='partial'):
        if not isinstance(data_payload, dict):  # TODO: Remove when responses are mapped to callbacks
            return

        if last_data_size is not None and last_data_size < len(self.data):
            raise IndexError('Overwriting of previous data prevented.')

        if response_type == 'full':
            self.data = Data(**data_payload)
            self.last_data_size = 0
        elif response_type == 'partial':
            self.data.extend(Data(**data_payload))
        else:
            raise ValueError()

        if len(data_payload['positions']):
            # stash the length of new data early to avoid events getting confused when pausing this thread
            old_last_data_size, self.last_data_size = self.last_data_size, len(self.data)
            invoke_as_event(self.update_graphs, self.data, old_last_data_size)

    def refresh_state(self, _, __):
        self.message_queue.queue.clear()
        self.message_queue.put(GetParametersRequest())
        self.message_queue.put(PullGraphsRequest())

    def update_graphs(self, data, last_data_size):
        self.graph_manager_widget.update_graphs(data, last_data_size)
        # x, y = zip(*data.positions)
        #
        # for metric_name in data.metrics:
        #     self._update_graph(metric_name, x, y, data.metrics[metric_name])
        #
        # self._update_graph('score', x, y, data.scores)
        # self._update_graph('variance', x, y, data.variances)

    def set_graphs(self, graphs):
        self.graph_manager_widget.set_graphs(graphs, self.data)

    def subscribe(self, callback, response_type: Union[Type[Message], None] = None, invoke_as_event: bool = False):
        self.callbacks[response_type].append((callback, invoke_as_event))

    def unsubscribe(self, callback, response_type: Union[Type[Message], None] = None):
        self.callbacks[response_type] = list(filter(lambda match_callback, invoke_as_event: match_callback == callback, self.callbacks[response_type]))

    def open_data(self):
        name, filter = QFileDialog.getOpenFileName(filter=("YAML (*.yml)"))
        if not name:
            return

        if len(self.data):
            result = QMessageBox.question(self,
                                          'Clear current data?',
                                          "Loading data will clear the current data set. Would you like to proceed?",
                                          buttons=QMessageBox.StandardButtons(QMessageBox.Yes | QMessageBox.Cancel),
                                          defaultButton=QMessageBox.Yes)
            if result != QMessageBox.Yes:
                return

        self.data = Data(**load(open(name, 'r'), Loader=Loader))
        self.last_data_size = len(self.data)
        # self.graph_manager_widget.reset()
        self.message_queue.put(PullGraphsRequest())
        # self.update_graphs(self.data, 0)
        if self.state_manager_widget.state == CoreState.Connecting:
            logger.warning('Data has been loaded before connecting to an experiment server. Remember to reload data after a connection is established.')
        else:
            result = QMessageBox.question(self,
                                          'Send data to server?',
                                          "Would you like to send the opened data to the experiment server? This will overwrite the server's current data.",
                                          buttons=QMessageBox.StandardButtons(QMessageBox.Yes | QMessageBox.No),
                                          defaultButton=QMessageBox.Yes)
            if result == QMessageBox.Yes:
                self.message_queue.put(PushDataRequest(self.data.as_dict()))

    def save_data(self):
        name, filter = QFileDialog.getSaveFileName(filter=("YAML (*.yml)"))
        if not name:
            return
        with self.data.r_lock():
            dump(self.data.as_dict(), open(name, 'w'), Dumper=Dumper)

    def open_parameters(self):
        name, filter = QFileDialog.getOpenFileName(filter=("YAML (*.yml)"))
        if not name:
            return
        state = load(open(name, 'r'), Loader=Loader)
        self.configuration_widget.parameter.restoreState(state, addChildren=True, removeChildren=True)

    def save_parameters(self):
        name, filter = QFileDialog.getSaveFileName(filter=("YAML (*.yml)"))
        if not name:
            return
        state = self.configuration_widget.parameter.saveState(filter='user')
        dump(state, open(name, 'w'), Dumper=Dumper)

    def new_data(self):
        if len(self.data):
            result = QMessageBox.question(self,
                                          'Clear current data?',
                                          "There is an active data set in memory. Would you like to proceed with clearing the current data?",
                                          buttons=QMessageBox.StandardButtons(QMessageBox.Yes | QMessageBox.Cancel),
                                          defaultButton=QMessageBox.Yes)
            if result != QMessageBox.Yes:
                return

        self.data = Data()
        self.graph_manager_widget.reset()

    def close_zmq(self):
        if self.update_thread.running:
            logger.info('waiting for update thread to finish')
            self.update_thread.requestInterruption()
            self.update_thread.wait()
        if self.socket:
            logger.debug('Closing socket')
            self.socket.close()
            self.socket = None
        if self.context:
            logger.debug('Closing context')
            self.context.term()
            self.context = None

    def closeEvent(self, event):
        if not self.close_demo(confirm=True):
            event.ignore()
            return
        if self.data and len(self.data):
            result = QMessageBox.question(self,
                                      'Save data?',
                                      "You have unsaved data in the active experiment. Do you want to save the data?",
                                      buttons=QMessageBox.StandardButtons(QMessageBox.Yes | QMessageBox.No | QMessageBox.Cancel),
                                      defaultButton=QMessageBox.Yes)
            if result == QMessageBox.Yes:
                self.save_data()
            if result in [QMessageBox.Yes, QMessageBox.No]:
                event.accept()
                self.close_zmq()
            else:
                event.ignore()
        else:
            event.accept()
            self.close_zmq()

    def start_demo(self, demo_key):
        # If not communicating with localhost, dump connection to server
        if self.core_address != 'localhost':
            self.core_address = 'localhost'
            self.init_socket()

        # if there's a child process server, exit it
        if not self.close_demo(confirm=True):
            return

        suffix = Path(sys.executable).suffix 
        demo_exe = (Path(sys.executable).parent/'tsuchinoko_demo').with_suffix(suffix if suffix=='.exe' else '')
        # print(demo_exe)
        self._server = subprocess.Popen([demo_exe, demo_key])

    def start_server(self, path):
        suffix = Path(sys.executable).suffix
        bootstrap_exe = (Path(sys.executable).parent / 'tsuchinoko_bootstrap').with_suffix(suffix if suffix == '.exe' else '')
        # print(bootstrap_exe)
        self._server = subprocess.Popen([bootstrap_exe, path])

    def close_demo(self, confirm=False):
        if self._server:
            if confirm:
                result = QMessageBox.question(self,
                                              'Shutdown server?',
                                              "A Tsuchinoko experiment server is currently running. Would you like to stop the server?",
                                              buttons=QMessageBox.StandardButtons(
                                                  QMessageBox.Yes | QMessageBox.Cancel),
                                              defaultButton=QMessageBox.Yes)
                if result != QMessageBox.Yes:
                    return False
            self.message_queue.put(ExitRequest())
            try:
                self._server.wait(3)

            except subprocess.TimeoutExpired:
                result = QMessageBox.question(self,
                                              'Server not responding.',
                                              "The currently running demo server is not responding. Would you like to terminate it?",
                                              buttons=QMessageBox.StandardButtons(
                                                  QMessageBox.Yes | QMessageBox.Cancel),
                                              defaultButton=QMessageBox.Yes)
                if result == QMessageBox.Yes:
                    self._server.kill()
                    return True
                elif result == QMessageBox.Cancel:
                    return False
        return True

    stop_server = close_demo

    def open_server_editor(self):
        self.editor_window = ServerEditor(main_window=self)
        self.editor_window.show()



import numpy as np
import pyqtgraph
from pyqtgraph import ImageView


# Fix for reaching last frame in ImageView timeline
class ImageView(ImageView):
    def timeIndex(self, slider):
        ## Return the time and frame index indicated by a slider
        if self.image is None:
            return (0, 0)

        t = slider.value()

        if not hasattr(self, "tVals"):
            return (0, 0)

        xv = self.tVals
        if xv is None:
            ind = int(t)
        else:
            if len(xv) < 2:
                return (0, 0)
            inds = np.argwhere(xv <= t)  # <- The = is import to reach the last value
            if len(inds) < 1:
                return (0, t)
            ind = inds[-1, 0]
        return ind, t

    def setCurrentIndex(self, ind):
        super(ImageView, self).setCurrentIndex(ind)
        (ind, time) = self.timeIndex(self.timeLine)
        self.sigTimeChanged.emit(ind, time)


pyqtgraph.__dict__["ImageView"] = ImageView


